/*-----------------------------------------------------------------------------
 *-----------------------------------------------------------------------------
 *- File: K4Board.c 
 *- Author: Owen Lin
 *- Last Modified: 7-8-97
 *- Description: This file contains the implementations of all the functions
 *-              defined in the header file K4Board.h. They implement all of
 *-              the operations defined for a konnekt4 board needed to run a
 *-              game.
 *-
 *-----------------------------------------------------------------------------
 *-----------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "K4Board.h"

/* recursions is used for development purposes, to count number of function
   calls */
int recursions;

/*-----------------------------------------------------------------------------
 * InitGame
 * --------
 * Description: This function initializes the data structures of the main game 
 *              which consist of the struct main_board, and the int turn.
 * Preconditions: turn, and main_board have been properly declared 
 * Postconditions: turn and all elements of main_board are initialized to 
 *                 proper start game values  
 *-----------------------------------------------------------------------------
 */
void InitGame() {

  int i;
  int j;

  /* reset to black's turn */
  turn = 0;
  /* set the number of chips */
  main_board.chips = 42;
  /* initialize chips_in_column to none, and game_board to -1 */
  for(i = 0; i < 9; i++) {
    main_board.chips_in_column[i] = 0;
    for(j = 0; j < 8; j++) {
      main_board.game_board[i][j] = -1;
    }
  }
  /* make all winning positins possible */
  for(i = 0; i < 2; i++) {
    for(j = 0; j < 69; j++) {
      main_board.win_valid[i][j] = 1;
      
    }
  }
}

/*-----------------------------------------------------------------------------
 * ToggleTurn
 * ----------
 * Description: This is a simple function. It is pretty much uneccessary. I
 *              included it to keep everything neat and tidy. It basically 
 * 	        switches the turn.
 * Preconditions: global named turn is declared
 * Postconditions: it is the other player's turn
 *-----------------------------------------------------------------------------
 */

void ToggleTurn() {
  turn = ! turn;
}

/*-----------------------------------------------------------------------------
 * IsDraw
 * ------
 * Description: This function returns 1 if the current game is a draw and 0
 *              otherwise.
 * Preconditions: none
 * Postconditions: none
 *-----------------------------------------------------------------------------
 */
int IsDraw(const struct K4Board* board) {
  return (board->chips == 0);
}

/*-----------------------------------------------------------------------------
 * IsColumnFull
 * ------
 * Description: This function returns 1 if the number of chips in the column
 *              is six, which is the maximum number of chips allowed.
 * Preconditions: none
 * Postconditions: none
 *-----------------------------------------------------------------------------
 */
int IsColumnFull(int column, const struct K4Board* board) {
  return (board->chips_in_column[column] == 6);
}

/*-----------------------------------------------------------------------------
 * ColumnHeight
 * ------
 * Description: This function returns the number of chips in the column
 * Preconditions: none
 * Postconditions: none
 *-----------------------------------------------------------------------------
 */
int ColumnHeight(int column, const struct K4Board* board) {
  return (board->chips_in_column[column]);
}

/*-----------------------------------------------------------------------------
 * CopyK4Board
 * ------
 * Description: This function is used to copy K4Board data structures
 * Preconditions: none.
 * Postconditions: The fields of *new_board are identical to those of
 *                 *board_to_copy.
 *-----------------------------------------------------------------------------
 */
void CopyK4Board(struct K4Board* new_board, 
		 const struct K4Board* board_to_copy) {
  int i;
  int j;
  
  /* copy the number of chips left */
  new_board->chips = board_to_copy->chips;
  /* copy the chips_in_column array and game_board in one shot */
  for(i = 0; i < 9; i++) {
    new_board->chips_in_column[i] = board_to_copy->chips_in_column[i];
    for(j = 0; j < 8; j++) {
      new_board->game_board[i][j] = board_to_copy->game_board[i][j];
    }
  }
  /* now copy the win_valid array */
  for(i = 0; i < 2; i++) {
    for(j = 0; j < 69; j++) {
      new_board->win_valid[i][j] = board_to_copy->win_valid[i][j];
    }
  }
  /* and last, the changes_length and last_changes array */
  new_board->changes_length = board_to_copy->changes_length;
  for(i = 0; i < new_board->changes_length; i++) {
    new_board->last_changes[i] = board_to_copy->last_changes[i];
  }
}

/*-----------------------------------------------------------------------------
 * PutChip
 * ------
 * Description: This function places a chip in the specified column of board.
 *              It marks off winning position possibility changes with a call
 *              to MarkPosition. If the column is full, no chips are placed.
 * Preconditions: none
 * Postconditions: If the column is not full, a chip is placed in it, and 
 *                 winning position possibilities are modified.
 *-----------------------------------------------------------------------------
 */
void PutChip(int player, int column, struct K4Board* board) {
  /* only add chip into column if it isn't full */
  if(board->chips_in_column[column] < 6) {
    /* increment the number of chips in the column */
    board->chips_in_column[column]++;
    /* decrement the number of chips left */
    board->chips--;
    /* put the player's chip in the column */
    board->game_board[column][board->chips_in_column[column]] = player;
    /* mark off all winning positions that are no longer possible */
    MarkPosition(player,column,board);    
  }
}

/*-----------------------------------------------------------------------------
 * RemoveChip
 * ------
 * Description: This function removes a chip from column, as long as it is not 
 *              empty. The winning positions made impossible by the chip that 
 *              was removed are restored as being possible. This function only
 *              works correctly for undoing the last move made.
 * Preconditions: none
 * Postconditions: The topmost chip has been removed from column. Possible
 *                 winning positions have been restored.
 *-----------------------------------------------------------------------------
 */
void RemoveChip(int column, struct K4Board* board) {
  
  int i;
  int other_player;
  
  /* only remove a chip if the column is not empty */
  if(board->chips_in_column[column] > 0) {
    /* set the value of the opposite player */
    other_player = ! board->game_board[column][board->chips_in_column[column]];
    /* clear the specified chip */
    board->game_board[column][board->chips_in_column[column]] = -1;
    /* decrement the number of chips in the column */
    board->chips_in_column[column]--;
    /* increment the number of chips left */
    board->chips++;
    /* last, restore the possible winning positions */
    for(i = 0; i < board->changes_length; i++) {
      board->win_valid[other_player][board->last_changes[i]] = 1;
    }
  }
}

/*-----------------------------------------------------------------------------
 * Winner
 * ------
 * Description: This function is used to determine if there was a win. It also
 *              computes the score of the current position. 
 * Preconditions: result is an integer array of size 2.
 * Postconditions: If there is a win, result[0] is set to the position in 
 *                 which the win occurred. Otherwise, it is a -1. result[1]
 *                 is always set to the score of position of the board.
 *-----------------------------------------------------------------------------
 */
void Winner(int player, int* result, const struct K4Board* board) {
  int i;
  int j;
  int temp_score;
  int test;
  int column;
  int row;
  div_t div_result;
  int win[4] = {0,0,0,0};

  temp_score = 0;  
  
  /* Basically, loop through all of the winning positions */
  for(i = 0; i < 69; i++) {
    /* examine them only if they are possible for the player */
    if(board->win_valid[player][i] == 1) {
      /* look at each coordinate in the winning position being examined */
      for(j = 0; j < 4; j++) {
	/* split the column and row coordinates */
	div_result = div(win_positions[i][j],10);
	column = div_result.quot;
	row = div_result.rem;	
	/* look at it if there is a chip at that coordinate */
	if(board->chips_in_column[column] >= row) {
	  /* if it is equal to the player mark that it is */
	  if(board->game_board[column][row] == player) {
	    win[j] = 1;
	  }
	  /* otherwise it isn't */
	  else {
	    win[j] = 0;
	  }
	}
	/* not occupied by player if there is no chip in that coordinate */
	else {
	  win[j] = 0;
	}
      }
      /* compute the score */
      test = win[0] + win[1] + win[2] + win[3];
      switch(test) {
      case 1 : 
	temp_score += 1;
	break;
      case 2 : 
	temp_score += 10;
	break;
      case 3 : 
	temp_score += 1000;
	break;
      case 4 : 
	/* if all four are the player's chips then a win has occurred */
	result[0] = i;
	result[1] = temp_score;
	return;
      }
    }
  }
  /* otherwise, no win has occurred */
  result[0] = -1;
  result[1] = temp_score;    
}

/*-----------------------------------------------------------------------------
 * MarkPosition
 * ------
 * Description: This function marks winning positions for the other player as
 *              impossible based on the most recent chip placed in column by
 *              player.
 * Preconditions: there is a chip in column of board
 * Postconditions: win_valid entries are marked as impossible for the other 
 *                 player.
 *-----------------------------------------------------------------------------
 */
void MarkPosition(int player, int column, struct K4Board* board) {
  
  int columnrow;
  int i;
  int j;
  int otherplayer;

  otherplayer = ! player;
  board->changes_length = 0;
  columnrow = (column * 10) + board->chips_in_column[column]; 
  for(i = 0; i < 69; i++) {
    if(board->win_valid[otherplayer][i] == 1) {
      for(j = 0; j < 4; j++) {     
	if(win_positions[i][j] == columnrow) {	 
	  board->win_valid[otherplayer][i] = 0;
	  board->last_changes[board->changes_length] = i;
	  board->changes_length++;
	}
      }
    }
  }  
}

void DFSMove(int player, int depth, float floor, int* result, 
	     const struct K4Board* board) {

  int column;
  int goodness;
  int best_goodness;
  int move; 
  int number_same_goodness;
  int DFS_result[2];
  int win_result[2];
  struct K4Board new_board;  
  
  recursions++;
  if(depth == 0) {
    result[0] = 0;
    result[1] = 0;
    return;
  }
  
  depth--;
  best_goodness = 0;
  goodness = 0;
  move = 0;
  number_same_goodness = 0;
  CopyK4Board(&new_board,board);

  for(column = 1; column < 8; column++) {    
    if(!IsColumnFull(column,&new_board)) {
      PutChip(player,column,&new_board); 
      Winner(player,win_result,&new_board);
      goodness = win_result[1];
      if(win_result[0] > -1) {
	result[0] = column;
	result[1] = goodness;
	return;
      }      
      if((float)goodness > floor) {     
	DFSMove(!player,depth,(10 * floor),DFS_result,&new_board);
	goodness += DFS_result[1];
	if(goodness > best_goodness) {
	  move = column;
	  best_goodness = goodness;
	  number_same_goodness = 1;
	}
	else if(goodness == best_goodness) {	
	  number_same_goodness++;
	  if((rand() % 10000) < ((float)1/(float)number_same_goodness) 
	     * 10000) {
	    move = column;
	    best_goodness = goodness;
	  }
	}	
      }
      RemoveChip(column,&new_board);
    }
  }
  result[0] = move;
  result[1] = best_goodness;
}

/* this function was adapted from code written by Keith Pomakis */
int DFSMove2(int player, int depth, const struct K4Board* board) {

  int column;
  int goodness;
  int best_goodness;
  int move;
  int win_result[2];
  int number_same_goodness;
  struct K4Board new_board;  
  
  number_same_goodness = 0;
  best_goodness = -(INT_MAX);
  goodness = 0;
  move = 0;
  CopyK4Board(&new_board,board);
 
  for(column = 1; column < 8; column++) {   
    if(!IsColumnFull(column,&new_board)) {
      PutChip(player,column,&new_board);     
      Winner(player,win_result,&new_board);    
      if(win_result[0] > -1) {        
	goodness = INT_MAX;
      }      
      else {
	goodness = DFSMove2Aux(player,depth,-(INT_MAX),-best_goodness,
			       &new_board);      
      }
      if(goodness > best_goodness) {
	move = column;
	best_goodness = goodness;
	number_same_goodness = 1;
      }
      else if(goodness == best_goodness) {	
	number_same_goodness++;
	if(rand()%10000 < ((float)1/(float)number_same_goodness) * 10000) {
	  move = column;
	}
      }	
      RemoveChip(column,&new_board);
    }
  }
  return move;
}


/* this function was adapted from code written by Keith Pomakis */
int DFSMove2Aux(int player, int depth, int alpha, int beta, 
	     const struct K4Board* board) {

  int column;
  int goodness;
  int best_goodness;
  int max_alpha_beta;
  int win_result[2];
  struct K4Board new_board;  
  
  depth--;
  recursions++;  

  if(depth == 0) {    
    Winner(player,win_result,board);   
    return win_result[1];
  }

  best_goodness = -(INT_MAX);
  goodness = 0;
  max_alpha_beta = alpha;
  CopyK4Board(&new_board,board); 
  
  for(column = 1; column < 8; column++) {    
    if(!IsColumnFull(column,&new_board)) {
      PutChip(!player,column,&new_board);  
      Winner(!player,win_result,&new_board);
      goodness = win_result[1];
      if(win_result[0] > -1) {
	goodness = INT_MAX - depth;
      }
      else {
	goodness = DFSMove2Aux(player,depth,-beta,-max_alpha_beta,
			       &new_board);     
      }
      if(goodness > best_goodness) {	
	best_goodness = goodness;
	if(best_goodness > max_alpha_beta) {
	  max_alpha_beta = best_goodness;
	}
      }    
      RemoveChip(column,&new_board);
    }
    if(best_goodness > beta) {
      break;
    }
  } 
  return -best_goodness;
}
